@@ -3,8 +3,13 @@
 
 use physfs_sys;
 use std::fmt;
+use std::sync::Mutex;
 
-static mut INSTANCED: bool = false;
+use lazy_static::lazy_static;
+
+lazy_static! {
+    static ref INSTANCED: Mutex<bool> = Mutex::new(false);
+}
 
 ///Represents an error returned by PhysicsFs
 pub struct PhysFsError {
@@ -82,13 +87,21 @@ unsafe impl Sync for PhysFsHandle {}
 
 impl std::io::Read for PhysFsHandle {
     fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
-        Ok(unsafe {
+        let read_bytes = unsafe {
             physfs_sys::PHYSFS_readBytes(
                 self.handle,
                 buf.as_mut_ptr() as *mut std::ffi::c_void,
                 buf.len() as u64,
-            ) as usize
-        })
+            )
+        };
+        if read_bytes == -1 {
+            Err(std::io::Error::new(
+                std::io::ErrorKind::Other,
+                get_last_error().err().unwrap(),
+            ))
+        } else {
+            Ok(read_bytes as usize)
+        }
     }
 }
 
@@ -185,13 +198,11 @@ impl PhysFs {
     ///When calling this method for the first time an instance of Self will be returned, if called
     ///again it will return None ensuring that no more than one instance exists
     pub fn get() -> Option<Self> {
-        //TODO: make thread safe
-        unsafe {
-            let ret = if !INSTANCED { Some(Self::new()) } else { None };
-            INSTANCED = true;
+        let mut instanced_lock = INSTANCED.lock().unwrap();
+        let ret = if !*instanced_lock { Some(Self::new()) } else { None };
+        *instanced_lock = true;
 
-            ret
-        }
+        ret
     }
 
     ///Mounts `dir` to `mount_point`.
@@ -241,6 +252,38 @@ impl PhysFs {
     pub fn open_append(&self, path: impl AsRef<str>) -> Result<PhysFsHandle, PhysFsError> {
         unsafe { self.make_handle(physfs_sys::PHYSFS_openAppend(to_cstr!(path))) }
     }
+
+    ///Enumerates the files in a given searchpath directory. None if erros occur
+    pub fn enumerate_files(&self, path: impl AsRef<str>) -> Option<Vec<String>> {
+        let mut list = unsafe {physfs_sys::PHYSFS_enumerateFiles(to_cstr!(path))};
+        if list == std::ptr::null_mut() {
+            return None;
+        }
+
+        let mut res = vec![];
+        while unsafe{*list} != std::ptr::null_mut() {
+            unsafe {
+                let filename = std::ffi::CStr::from_ptr(*list).to_str().unwrap();
+                res.push(filename.to_owned());
+            }
+            list = ((list as usize) + std::mem::size_of_val(&list)) as _;
+        }
+
+        return Some(res);
+    }
+
+    pub fn is_directory(&self, path: impl AsRef<str>) -> bool {
+        unsafe{
+            physfs_sys::PHYSFS_isDirectory(to_cstr!(path)) == 1
+        }
+    }
+}
+
+impl Drop for PhysFs {
+    fn drop(&mut self) {
+        let mut instanced_lock = INSTANCED.lock().unwrap();
+        *instanced_lock = false;
+    }
 }
 
 #[cfg(test)]
@@ -248,10 +291,19 @@ mod tests {
     use super::*;
     use std::io::*;
 
+    //tests seem to run in parallel threads, wait for the instance to be available
+    fn wait_instance() -> PhysFs {
+        let mut instance = None;
+        while instance.is_none() {
+            instance = PhysFs::get();
+        }
+        instance.unwrap()
+    }
+
     const TEST_TEXT: &'static str = "testing\n";
     #[test]
     fn it_works() {
-        let mut fs = PhysFs::get().unwrap();
+        let mut fs = wait_instance();
         fs.mount("alol.zip", "", true).err().unwrap(); //muest give none
         fs.mount("test.zip", "", true).unwrap();
         fs.mount("./", "", true).unwrap();
@@ -268,4 +320,16 @@ mod tests {
 
         assert!(PhysFs::get().is_none());
     }
+
+    #[test]
+    fn files_enumeration() {
+        let mut fs = wait_instance();
+        fs.mount("test.zip", "", true).unwrap();
+
+        let files = fs.enumerate_files("/").unwrap();
+        assert_eq!(vec!["test"], files);
+        assert!(!fs.is_directory("test"));
+        fs.mount("./", "/", true);
+        assert!(fs.is_directory("src"));
+    }
 }
